//  main.cpp
//  Chain Algorithm
//
//  Created by Lingxun Hu on 7/22/12.
//  Copyright (c) 2012 RPI. All rights reserved.

#include <iostream>
#include <algorithm>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <vector>
#include <fstream>
#include <time.h>
#define tmin 0.01
#define tmax 5.0
#include "Node.h"

using namespace std;

bool compare (const Node_type1& x, const Node_type1& y)// personally defined function compare, which compare two struct and put them in the order of time element

{
    return x.time < y.time;
}

void reading (vector<Node_type1>& x, vector<Node_type1>& y)//reading reads send and receive data from files into two vectors.
{
    ifstream in1("send.txt"), in2("receive.txt");
    Node_type1 node;
    
    while (in1 >> node.number >> node.time) 
    {
        x.push_back(node);
    }
    while (in2 >> node.number >> node.time) 
    {
        y.push_back(node);
    }
}




vector<Node_type2> chain(const vector<Node_type1>& x,const vector<Node_type1>& y)// find all possible matching using the chain algorithm mentioned in the hidden group paper.

{
    int i=0,j=0;
    vector<Node_type2> ret;
    Node_type2 temp;
    while (i!=x.size() && j!= y.size()) 
    {
        if ((y[j].time - x[i].time) < tmin) ++j;// the time intervel is too small, thus move the pointer of receiver forward
        else if ((y[j].time - x[i].time) > tmin && (y[j].time - x[i].time) < tmax)// time intervel is good, thus record this matching and move both pointers of sender and reveiver forward
        {
            if (x[i].number != y[j].number)// make sure the sender and receiver are not one person
            {
                temp.S = x[i].number;
                temp.R = y[j].number;
                temp.time = x[i].time;
                ret.push_back(temp);
                ++i;++j;
            }
            else 
            {
                if((rand()%2)) ++i;
                else ++j;
            }// if sender and receiver are actually one person, we generate a random number to decide to move pointer of sender or the pointer of receiver
        }    
        else ++i;// the time intervel is too big, thus move the pointer of sender forward
    }
    return ret;
}



void output(vector<Node_type2>& x)
{
    ofstream out("chain algorithm.txt");
    int i;
    for (i = 0; i != x.size();++i)out << x[i].S << "    "<< x[i].R << "     " << x[i].time << endl;
    cout << "The number of matchings is " << i << endl;
}

int main (int argc, const char * argv[])
{
    double t1 = clock();// recording the beginning time of the program
    vector<Node_type1> sender,receiver;
    
    reading(sender, receiver);// reads input data into two vectors sender and receiver
    
    sort(sender.begin(),sender.end(),compare);// sort sender in time order
    sort(receiver.begin(),receiver.end(),compare);//sort receiver in time order
    
    vector<Node_type2> matchings;
    matchings = chain(sender, receiver);// use chain algorithm to find all the possible matchings
    
    output(matchings);// write all matchings into a file
    double t2 = clock();// record ending time
    cout << "The running time is " << t2 - t1 << endl;// output running time of the program
    return 0;
}
