//
//  main.cpp
//  Dynamic Matching
//
//  Created by Lingxun Hu on 7/19/12.
//  Copyright (c) 2012 RPI. All rights reserved.
//

#include <iostream>
#include <algorithm>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <vector>
#include <fstream>
#include <time.h>
#include "Node.h"// the header file Node.h defines the Node_type1

#define tmax 5
#define tmin 0.01

using namespace std;

bool compare (const Node_type1& x, const Node_type1& y)// personally defined function compare, which compare two struct and put them in the order of time element
{
    return x.time < y.time;
}

void reading (vector<Node_type1>& x, vector<Node_type1>& y)//reading reads send and receive data from files into two vectors.
{
    ifstream in1("send.txt"), in2("receive.txt");
    Node_type1 node;
    
    while (in1 >> node.number >> node.time)//reads senders' identity numbers and sending times into the node's number and time parts one by one
    {
        x.push_back(node);//append to vector x
    }
    while (in2 >> node.number >> node.time) //similarly reads receivers' identity numbers and receiving times
    {
        y.push_back(node);//append to vector y
    }
}

void output(vector<Node_type2>& x)//writes the results into a file
{
    ofstream out("dynamic_matching.txt");
    int i;
    for (i = 0; i != x.size(); ++i)
    {
        out << x[i].S << "  " << x[i].R << "    "<< x[i].time <<endl;// writes the identity numbers of a pair of sender and receiver
    }
    cout <<"The matching number is " << i << endl;//output the total number of matchings we get at the end of the program
}

double Match_Weight(double t1, double t2)// determines the weight of a possible matching between a sender and a receiver
{
    double ret;//define the maximum and minimum time intervels between a pair of possile sender and receiver
    if ((t2 - t1) < tmax && (t2 - t1) > tmin) ret = 10.0 - 2.0 * (t2 - t1);// if a pair is possible we calculate the weight
    else ret = 0;// if a pair is not possible we set the weight to be 0
    return ret;
}

vector<Node_type2> Match_Causality(vector<Node_type1>& list1, vector<Node_type1>& list2)// matches the possible senders and receivers via dynamic programming method
{
    //char stores the direction when backtracking;
    //0 means change list2 index, 1 means change both, 2 means change list1 index
    pair<double, char> p;
    vector<pair<double, char> > v(list2.size(), p);
    
    //the double represents the weight of the match, the char stores the direction
    vector<vector<pair<double, char> > > match_grid(list1.size(), v);
    for(int i=0; i < list1.size(); i++)
    {
        for(int j=0; j < list2.size(); j++)
        {
            
            //Calculates left, bottom-left, and bottom weights, respectively  
            double weight1; double weight2; double weight3;
            if(j > 0){weight1 = match_grid[i][j-1].first;}
            else {weight1 = 0;}
            weight2 = Match_Weight(list1[i].time, list2[j].time);
            if(i > 0 && j > 0){weight2 += match_grid[i-1][j-1].first;}
            if(i > 0){weight3 = match_grid[i-1][j].first;}
            else {weight3 = 0;}
            
            //Finds maximum of all weights and stores that (along with the direction) in p
            double max_weight;
            if (list1[i].number != list2[j].number)//if it is not a loop, we make it a valid matching and calculate the weight
            {
                max_weight = max(weight1, weight2);
                max_weight = max(max_weight, weight3);
                p.first = max_weight;
            }
            else p.first = max(weight1,weight3);// if it is a loop, we think it is not possible and seek other possible matchings
            if(weight1 == max_weight){p.second = 0;}// we record every step we make after choosing the direction
            else if(weight2 == max_weight){p.second = 1;}
            else {p.second = 2;}
            match_grid[i][j] = p;// we store the path in a matrix
        }
    }
    
    //Traces back through all the pointers to find total matches.
    int i=list1.size() - 1; int j=list2.size() - 1;// i and j are at the ends of list1 and list2
    vector<Node_type2> matching;// matching is our final results
    while(i>=0 && j>=0)
    {
        char direction = match_grid[i][j].second;
        if (direction == 0){j--;}// trace back 
        else if(direction == 1)
        {
            Node_type2 times;
            times.S = list1[i].number; times.R = list2[j].number;times.time = list1[i].time;
            matching.push_back(times);
            i--;j--;
        }
        else{i--;}
    }
    double score = (match_grid[list1.size()-1][list2.size()-1]).first;// calculates the maximum weight
    cout << "The maximum weight is " << score << endl;
    return matching;
}



int main (int argc, const char * argv[])
{
    double t1 = clock();//recording the starting time
    vector<Node_type1> sender,receiver;//variables for reading the data
    reading(sender, receiver);// calling the reading function to read the input data into vectors
    
    sort(sender.begin(), sender.end(),compare);// using built-in sort function to sort the sender vector in ascending order of time 
    sort(receiver.begin(), receiver.end(),compare);// similarly sort the receiver vector in ascending order of time
    
    vector<Node_type2> matchings;// define vector matchings
    matchings = Match_Causality(sender, receiver);//function Match_Causality uses dynamic porgramming method to find all the possible matchings while avoiding the loops
    output(matchings);// output matchings write into file
    double t2 = clock();// recording the ending time 
    cout << "The running time is " << t2 - t1 << endl;//output the running time of the program
    return 0;
}


